from typing import Any

import click

from eopf import EOContainer, EOZarrStore, OpeningMode
from eopf.cli.cli import EOPFPluginCommandCLI, EOPFPluginGroupCLI
from eopf.logging import EOLogging
from eopf.qualitycontrol.eo_qc_processor import EOQCProcessor


class EOCLIQualityControlCheckTrigger(EOPFPluginCommandCLI):
    """EOTrigger cli command to apply check on a product.

    Parameters
    ----------
    context_settings: dict, optional
        default values provide to click

    See Also
    --------
    click.Command
    """

    name = "check"
    cli_params: list[click.Parameter] = [
        click.Argument(["product_path"], type=click.Path()),
        click.Option(["--output-product"], type=click.Path()),
        click.Option(["--output-report"], type=click.Path()),
        click.Option(["--config-folder"], type=click.Path()),
        click.Option(["--add-config-folder"], type=click.Path()),
    ]
    help = (
        "Request the OLQC check on zarr eoproduct\n\n\n"
        "Args:\n\n"
        "  product_path = Path to the zarr eoproduct\n\n"
        "  [--output-product] = Product to write the result to, if not not written\n\n"
        "  [--output-report] = Report to write to, if not not written\n\n"
        "  [--config-folder] = Base config folder,  if not default used\n\n"
        "  [--add-config-folder] = Additional config folder to use\n\n"
    )

    @staticmethod
    def callback_function(
        product_path: str,
        output_product: None | str,
        output_report: None | str,
        config_folder: None | str,
        add_config_folder: None | str,
    ) -> Any:
        """

        Parameters
        ----------
        product_path
        output_product
        output_report
        config_folder
        add_config_folder

        Returns
        -------

        """

        logger = EOLogging().get_logger()
        logger.info(f"Checking product  {product_path}")
        if output_product is None and output_report is None:
            logger.error("Need at least one of output_product or output_report")
            return 1
        config_folder = "default" if config_folder is None else config_folder
        eoqc_processor = EOQCProcessor(
            config_folder=config_folder,
            update_attrs=(output_product is not None),
            additional_config_folders=[add_config_folder] if add_config_folder is not None else None,
            report_path=output_report,
        )
        inproduct = EOZarrStore(url=product_path).load()
        if isinstance(inproduct, EOContainer):
            logger.error("Can't check an EOContainer")
            return 1
        eoqc_processor.check(inproduct)
        if output_product is not None:
            with EOZarrStore(url=output_product).open(mode=OpeningMode.CREATE) as store:
                store[""] = inproduct

        return 0


class EOCLIQualityControlTrigger(EOPFPluginGroupCLI):
    """EOTrigger cli command aggregator to quality control triggers

    Parameters
    ----------
    **attrs: Any
        any argument for click.Command, click.MultiCommand

    See Also
    --------
    click.Group
    """

    name = "qualitycontrol"
    cli_commands: list[click.Command] = [
        EOCLIQualityControlCheckTrigger(),
    ]
    help = "CLI commands to run EOQC processor"
