from typing import Any

import xarray as xr
import zarr
from xarray.core.datatree import DataTree

from eopf.common.file_utils import AnyPath


def create_dtree_from_zarr_group(zarr_group: zarr.Group, name: str, **kwargs: Any) -> DataTree:
    """Constructs a DataTree from a given Zarr group

    Parameters
    ----------
    zarr_group: zarr.Group
        The root Zarr group from which the DataTree will be constructed.
        A Zarr group is a collection of arrays and nested groups.
    name: str
        The name to be assigned to the root node of the DataTree.
    kwargs: Any
        Additional keyword arguments passed to xr.open_dataset

    Returns
    -------
    DataTree
        The constructed DataTree object with nodes corresponding to datasets and groups within the Zarr hierarchy.
    """

    def create_tree_from_group(group: zarr.Group, node_path: AnyPath, **kwargs: Any) -> DataTree:
        """Recursive function that performs the construction of the DataTree by traversing the Zarr group hierarchy.

        Parameters
        ----------:
        group: zarr.Group
            The current Zarr group being processed
        path: str, optional
            The path within the Zarr hierarchy for the current group

        Returns
        -------
        DataTree:
            A subtree corresponding to the current Zarr group, with child nodes for each nested group or dataset.
        """
        try:
            ds = xr.open_dataset(group.store, group=group.path, engine="zarr", **kwargs)
        except ValueError:  # Handle the case where the group does not contain a dataset
            ds = xr.Dataset()

        node: DataTree = DataTree(
            name=node_path.basename if node_path.original_url else name,
            dataset=ds,
        )

        for key, value in group.groups():
            node_subpath = (
                AnyPath(f"{node_path.original_url}{node_path.sep}{key}") if node_path.original_url else AnyPath(key)
            )
            child_node = create_tree_from_group(value, node_subpath, **kwargs)
            node[key] = child_node

        return node

    root_node_path = AnyPath("")
    return create_tree_from_group(zarr_group, root_node_path, **kwargs)


def open_datatree(produt_anypath: AnyPath, product_id: str, **kwargs: Any) -> DataTree:
    """Open and decode a EOPF-like Zarr product

    Parameters
    ----------
    produt_anypath: AnyPath
        Path to directory in file system or name of zip file (as an AnyPath object)
    product_id: str
        Id of the product as passed in the payload

    Returns
    -------
        DataTree
    """

    zarr_store = produt_anypath._fs.get_mapper(produt_anypath.path)
    # Open the Zarr root group
    root_group = zarr.open_group(zarr_store, mode="r")

    # Create the DataTree from the Zarr root group
    data_tree = create_dtree_from_zarr_group(root_group, product_id, **kwargs)

    return data_tree
