import importlib
from enum import Enum
from pathlib import Path
from typing import Any, Optional, Union

import dask.config
from dask.distributed import Client
from distributed import get_client, performance_report
from importlib_metadata import PackageNotFoundError

from eopf.exceptions import TriggeringConfigurationError
from eopf.logging.log import EOLogging


class ClusterType(Enum):
    LOCAL = "local"
    SSH = "ssh"
    KUBERNETES = "kubernetes"
    PBS = "pbs"
    SGE = "sge"
    LSF = "lsf"
    SLURM = "slurm"
    YARN = "yarn"
    GATEWAY = "gateway"
    ADDRESS = "address"
    CUSTOM = "custom"


def get_enum_from_value(value: str) -> ClusterType:
    for member in ClusterType:
        if member.value == value:
            return member
    raise ValueError("Value not found in the enum")


class DaskContext:
    """Representation of a dask context used to run
    dask with the correct configuration.

    Parameters
    ----------
    cluster_type: type of cluster to use
        can be one of:

            * **None**: don't create a new cluster, just connect to one
            * **local**: configure a :py:class:`~dask.distributed.LocalCluster`
            * **ssh**: configure a :py:func:`~dask.distributed.SSHCluster`
            * **kubernetes**: configure a :py:class:`~dask_kubernetes.KubeCluster`
            * **pbs**: configure a :py:class:`~dask_jobqueue.PBSCluster`
            * **sge**: configure a :py:class:`~dask_jobqueue.SGECluster`
            * **lsf**: configure a :py:class:`~dask_jobqueue.LSFCluster`
            * **slurm**: configure a :py:class:`~dask_jobqueue.SLURMCluster`
            * **slurm**: configure a :py:class:`~dask_jobqueue.SLURMCluster`
            * **yarn**: configure a :py:class:`~dask_yarn.YarnCluster`
            * **gateway**: configure a :py:class:`~dask_gateway.GatewayCluster`
            * **address**: to pass simple cluster address in the addr parameter
            * **custom**: to use Custom cluster class specified by following element in **cluster_config** element.

                - **module**: python path to module containing custom cluster
                - **cluster**: cluster class name
    address: str
        only for **str** cluster_type, specified cluster address to join.
    cluster_config: dict
        key value pairs of parameters to give to the cluster constructor
    client_config: dict
        element to configure :py:class:`~dask.distributed.Client`
    performance_report_file : Optional[str]
        path to report file
    """

    def __init__(
        self,
        cluster_type: Optional[ClusterType] = None,
        address: Optional[str] = None,
        cluster_config: Optional[dict[str, Any]] = None,
        client_config: Optional[dict[str, Any]] = None,
        performance_report_file: Optional[Union[str, Path]] = None,
    ) -> None:
        self._cluster: Optional[Any] = None
        self._logger = EOLogging().get_logger("eopf.daskconfig.dask_context_manager")
        self._client: Optional[Client] = None
        self._performance_report_file = performance_report_file
        self._performance_report: Optional[performance_report] = None
        # See https://distributed.dask.org/en/latest/diagnosing-performance.html
        if self._performance_report_file:
            self._performance_report = performance_report(filename=self._performance_report_file)
            self._logger.info(f"Performance report file requested : {self._performance_report_file}")

        self._client_config: dict[str, Any] = client_config if client_config else {}
        self._cluster_config: dict[str, Any] = cluster_config if cluster_config else {}

        self._logger.info(
            f"Initialising an {cluster_type} cluster with client conf : {client_config} "
            f"and cluster config {cluster_config}",
        )

        # detect any other client, this might cause conflict on the cluster
        try:
            client = get_client()
            # if a client exist log its infos
            if client is not None:
                self._logger.warning(f"A Dask client is already active: {id(client)} : {client}")
        except Exception:
            self._logger.debug("No Dask client found")

        # setup cluster
        if cluster_type == ClusterType.LOCAL:
            """
            Starts a local cluster. Options from dask documentation
            https://distributed.dask.org/en/latest/api.html#distributed.LocalCluster:
            name=None, n_workers=None, threads_per_worker=None, processes=None, loop=None, start=None, host=None,
            ip=None, scheduler_port=0, silence_logs=30, dashboard_address=':8787', worker_dashboard_address=None,
            diagnostics_port=None, services=None, worker_services=None, service_kwargs=None, asynchronous=False,
            security=None, protocol=None, blocked_handlers=None, interface=None, worker_class=None,
            scheduler_kwargs=None, scheduler_sync_interval=1, **worker_kwargs

            """
            from dask.distributed import LocalCluster

            self._cluster = LocalCluster(**self._cluster_config)
        elif cluster_type == ClusterType.SSH:
            from dask.distributed import SSHCluster

            self._cluster = SSHCluster(**self._cluster_config)
        elif cluster_type == ClusterType.KUBERNETES:
            try:
                from dask_kubernetes import KubeCluster

                self._cluster = KubeCluster(**self._cluster_config)
            except ModuleNotFoundError:
                raise PackageNotFoundError("Package dask_kubernetes should be installed.")
        elif cluster_type == ClusterType.PBS:
            try:
                from dask_jobqueue import PBSCluster

                self._cluster = PBSCluster(**self._cluster_config)
            except ModuleNotFoundError:
                raise PackageNotFoundError("Package dask_jobqueue should be installed.")
        elif cluster_type == ClusterType.SGE:
            try:
                from dask_jobqueue import SGECluster

                self._cluster = SGECluster(**self._cluster_config)
            except ModuleNotFoundError:
                raise PackageNotFoundError("Package dask_jobqueue should be installed.")
        elif cluster_type == ClusterType.LSF:
            try:
                from dask_jobqueue import LSFCluster

                self._cluster = LSFCluster(**self._cluster_config)
            except ModuleNotFoundError:
                raise PackageNotFoundError("Package dask_jobqueue should be installed.")
        elif cluster_type == ClusterType.SLURM:
            try:
                from dask_jobqueue import SLURMCluster

                self._cluster = SLURMCluster(**self._cluster_config)
            except ModuleNotFoundError:
                raise PackageNotFoundError("Package dask_jobqueue should be installed.")
        elif cluster_type == ClusterType.YARN:
            try:
                from dask_yarn import YarnCluster

                self._cluster = YarnCluster(**self._cluster_config)
            except ModuleNotFoundError:
                raise PackageNotFoundError("Package dask_yarn should be installed.")
        elif cluster_type == ClusterType.GATEWAY:
            self._setup_gateway_cluster()
        elif cluster_type == ClusterType.ADDRESS:
            if address is None:
                raise TriggeringConfigurationError("addr parameter is mandatory for STR cluster connexion")
            if "address" not in self._client_config:
                self._client_config["address"] = address
        elif cluster_type == ClusterType.CUSTOM:
            self._setup_custom_cluster()
        elif cluster_type is None:
            # No cluster type provided but an address is given
            if address is not None and "address" not in self._client_config:
                self._client_config["address"] = address
        else:
            raise TriggeringConfigurationError("Unhandled dask context configuration")
        self._logger.info(f"DASK Cluster : {str(self._cluster)}")

    def _setup_custom_cluster(self) -> None:
        module_name: str = "NotFound"
        try:
            module_name = self._cluster_config.pop("module")
            cluster_class_name = self._cluster_config.pop("cluster")
            cluster = getattr(importlib.import_module(module_name), cluster_class_name)(
                **self._cluster_config,
            )
            self._cluster = cluster
        except ModuleNotFoundError:
            raise ModuleNotFoundError(f"Module {module_name} not found, corresponding package should be installed")

    def _setup_gateway_cluster(self) -> None:
        try:
            from dask_gateway import Gateway
            from dask_gateway.auth import get_auth

            # setup gateway Auth
            # one of ("kerberos", "jupŷterhub", "basic") or a python pass to the auth class
            auth_kwargs = self._cluster_config.pop("auth", {})
            workers = self._cluster_config.pop("workers", 1)
            auth_type = auth_kwargs.pop("type")
            dask.config.set(gateway__auth__kwargs=auth_kwargs)
            auth = get_auth(auth_type)
            # setup gateway
            gateway_url = self._cluster_config.pop("address")
            gateway = Gateway(address=gateway_url, auth=auth)
            available_clusters = gateway.list_clusters()
            # reuse existing cluster ?
            previous_cluster_name = self._cluster_config.pop("reuse_cluster", None)
            if previous_cluster_name is not None and any(
                previous_cluster_name == cluster.name for cluster in available_clusters
            ):
                self._logger.info(f"Reusing previous cluster {previous_cluster_name}")
                self._logger.debug(
                    f"Previous cluster status page : {gateway_url}/clusters/{previous_cluster_name}/status",
                )
                cluster = gateway.connect(previous_cluster_name)
            else:
                self._logger.info(f"Creating new cluster with config {self._cluster_config}")
                cluster = gateway.new_cluster(**self._cluster_config)
            # scale the cluster
            if isinstance(workers, int):
                cluster.scale(workers)
            elif isinstance(workers, dict):
                cluster.adapt(**workers)
            self._cluster = cluster
        except ModuleNotFoundError:
            raise PackageNotFoundError("Package dask_gateway should be installed.")

    def __del__(self) -> None:
        if self._client is not None:
            self._client.close()
        if self._cluster is not None:
            self._cluster.close()
        self._cluster = None
        self._client = None
        self._performance_report = None

    def __enter__(self) -> "DaskContext":
        self._logger.debug(f"Starting dask client and cluster {self}")
        if self._cluster is not None:
            self._cluster.__enter__()
            self._logger.debug(f"Starting dask cluster : {id(self._cluster)}")
            self._client = Client(address=self._cluster, **self._client_config)
        else:
            self._client = Client(**self._client_config)
        self._logger.info(f"DASK Client : {str(self._client)}")
        self._logger.debug(f"Starting dask client : {id(self._client)}")
        self._client.__enter__()

        if self._performance_report:
            self._performance_report.__enter__()
        return self

    def __exit__(self, *args: Any, **kwargs: Any) -> None:
        self._logger.debug(f"Stopping dask client and cluster {self}")
        if self._performance_report:
            self._performance_report.__exit__(*args, **kwargs)

        if self._client is not None:
            self._logger.debug(f"Stopping dask client : {id(self._client)}")
            self._logger.debug(f"processing : {self._client.processing()}")
            self._client.__exit__(*args, **kwargs)

        if self._cluster is not None:
            self._logger.debug(f"Stopping dask cluster : {id(self._cluster)}")
            self._cluster.__exit__(*args, **kwargs)

        self._client = None
        self._cluster = None

    @property
    def cluster(self) -> Optional[Any]:
        return self._cluster

    @property
    def client(self) -> Optional[Client]:
        return self._client

    def __str__(self) -> str:
        return f"cluster : {self.cluster}, client : {self.client}"
