from datetime import datetime
from typing import Any

import numpy
from pandas import Timedelta, to_datetime
from pytz import UTC

from eopf.exceptions import FormattingError

from .abstract import EOAbstractFormatter


class ToUNIXTimeSLSTRL1(EOAbstractFormatter):
    """Formatter for unix time conversion for SLSTR L1 ANX time and calibration time variables"""

    # docstr-coverage: inherited
    name = "to_unix_time"

    def _format(self, input: Any) -> int:
        """Convert input to unix time

        Parameters
        ----------
        input: Any

        Returns
        ----------
        eov: EOVariable
            EOVariable with the data converted to unix time

        Raises
        ----------
        FormattingError
            When formatting can not be performed
        """
        try:
            # compute the start and end times
            start = to_datetime(datetime.fromtimestamp(0, tz=UTC))
            end = to_datetime(input[:], utc=True)

            # compute and convert the time difference into microseconds
            time_delta = numpy.int64((end - start) // Timedelta("1microsecond"))

            # convert to int since neither datetime nor numpy int types are serialisable
            return int(time_delta)
        except Exception as e:
            raise FormattingError(f"{e}") from e


class ToISO8601(EOAbstractFormatter):
    """Formatter for ISO8601 (time) conversion"""

    # docstr-coverage: inherited
    name = "to_ISO8601"

    def _format(self, input: str) -> str:
        """Convert time to ISO8601 standard, e.g: 20220506T072719 -> 2022-05-06T07:27:19Z

        Parameters
        ----------
        input: str
            xpath

        Returns
        ----------
        date_string: strftime (string-like time format)
            String containing date converted to ISO standard.

        Raises
        ----------
        FormattingError
            When formatting can not be performed
        """
        try:
            dt_obj = datetime.strptime(input, "%Y%m%dT%H%M%S")
            date_string = dt_obj.strftime("%Y-%m-%dT%H:%M:%SZ")
            return date_string
        except ValueError:
            raise FormattingError(f"Input {input} cannot be converted to ISO standard.")
