from abc import ABC, abstractmethod
from dataclasses import dataclass

from eopf.product import EOProduct
from eopf.qualitycontrol.eo_qc_utils import EOQCFormulaEvaluator


@dataclass(frozen=True)
class EOQCPartialCheckResult:
    """
    Check result partial output of _check
    """

    status: bool = True
    message: str = ""

    def __add__(self, other: "EOQCPartialCheckResult") -> "EOQCPartialCheckResult":
        if not isinstance(other, EOQCPartialCheckResult):
            raise TypeError("Operands must be an EOQCPartialCheckResult")
        if self.message != "":
            return EOQCPartialCheckResult(self.status and other.status, f"{self.message} ; {other.message}".strip())
        else:
            return EOQCPartialCheckResult(self.status and other.status, f"{other.message}".strip())


@dataclass(frozen=True)
class EOQCCheckResult:
    """
    Check result
    """

    id: str
    version: str
    thematic: str
    description: str
    status: bool
    message: str

    @classmethod
    def from_partial(
        cls,
        partial_result: EOQCPartialCheckResult,
        id: str,
        version: str,
        thematic: str,
        description: str,
    ) -> "EOQCCheckResult":
        return EOQCCheckResult(
            id=id,
            version=version,
            thematic=thematic,
            description=description,
            message=partial_result.message,
            status=partial_result.status,
        )


@dataclass
class EOQC(ABC):
    """Quality check class.

    Parameters
    ----------
    id: str
        The identifier of the quality check.
    version: str
        The version of the quality check in format XX.YY.ZZ .
    thematic: str
        Thematic of the check RADIOMETRIC_QUALITY/GEOMETRIC_QUALITY/GENERAL_QUALITY...
    description: str
        Simple description of the check (less than 100 chars)
    precondition: EOQCFormulaEvaluator
        Precondition evaluator

    Methods
    --------
    _check: EOQCPartialCheckResult
        Abstract method to be implemented by checks implementation



    """

    id: str
    version: str
    thematic: str
    description: str
    precondition: EOQCFormulaEvaluator

    def check(self, eoproduct: EOProduct) -> EOQCCheckResult:  # pragma: no cover
        """Check method for a quality check.

        Parameters
        ----------
        eoproduct: EOProduct
            The product to check.

        Returns
        -------
        bool
            Status of the quality check, true if it's ok, false if not.
        """
        if not bool(self.precondition.evaluate(eoproduct=eoproduct)):
            partial_result = EOQCPartialCheckResult(
                status=True,
                message=f"SKIPPED: Precondition evaluate False with formula : {self.precondition.formula}",
            )
        else:
            partial_result = self._check(eoproduct)

        # Create and return
        return EOQCCheckResult.from_partial(
            partial_result=partial_result,
            id=self.id,
            version=self.version,
            thematic=self.thematic,
            description=self.description,
        )

    @abstractmethod
    def _check(self, eoproduct: EOProduct) -> EOQCPartialCheckResult:  # pragma: no cover
        """Check method for a quality check.

        Parameters
        ----------
        eoproduct: EOProduct
            The product to check.

        Returns
        -------
        EOQCPartialCheckResult
            Status of the quality check, and the result message
        """

        raise NotImplementedError
