from abc import abstractmethod
from dataclasses import dataclass
from typing import Any

import numpy as np

from eopf import EOGroup, EOProduct, EOVariable
from eopf.qualitycontrol.eo_qc import EOQC, EOQCPartialCheckResult
from eopf.qualitycontrol.eo_qc_factory import EOQCFactory


@dataclass
class EOQCVarBase(EOQC):
    """
     This EOQC is the base of all variables checks.

    Parameters
     ----------
     id: str
         The identifier of the quality check.
     version: str
         The version of the quality check in format XX.YY.ZZ .
     thematic: str
         Thematic of the check RADIOMETRIC_QUALITY/GEOMETRIC_QUALITY/GENERAL_QUALITY...
     description: str
         Simple description of the check (less than 100 chars)
     precondition: EOQCFormulaEvaluator
         Precondition evaluator
     variables: list[str]
         List of the variables to check



    """

    variables: list[str]

    # docstr-coverage: inherited
    def _check(self, eoproduct: EOProduct) -> EOQCPartialCheckResult:
        res = EOQCPartialCheckResult()
        if len(self.variables) == 0:
            return EOQCPartialCheckResult(status=True, message="SKIPPED: No variable/group to check")
        for var_path in self.variables:
            try:
                attr_value = eoproduct[var_path]
            except KeyError:
                res = res + EOQCPartialCheckResult(
                    status=False,
                    message=f"FAILED : Variable/Group {var_path} is not available "
                    f"in product for check {type(self).__name__}",
                )
                continue
            res = res + self._var_check(var_path, attr_value, eoproduct)
        if res is None:
            return EOQCPartialCheckResult(status=False, message="ERROR: This should not happen")
        return res

    @abstractmethod
    def _var_check(self, var_path: str, value: Any, eoproduct: EOProduct) -> EOQCPartialCheckResult:
        pass


@dataclass
@EOQCFactory.register_eoqc("var_exists")
class EOQCVarAvailable(EOQCVarBase):
    """
     This EOQC checks that the variables/groups are available in the product.

    Parameters
     ----------
     id: str
         The identifier of the quality check.
     version: str
         The version of the quality check in format XX.YY.ZZ .
     thematic: str
         Thematic of the check RADIOMETRIC_QUALITY/GEOMETRIC_QUALITY/GENERAL_QUALITY...
     description: str
         Simple description of the check (less than 100 chars)
     precondition: EOQCFormulaEvaluator
         Precondition evaluator
     variables: list[str]
         List of the variables to check


    """

    def _var_check(self, var_path: str, value: Any, eoproduct: EOProduct) -> EOQCPartialCheckResult:
        if value is not None:
            return EOQCPartialCheckResult(
                status=True,
                message=f"PASSED : Variable/Group {var_path} is available in " f"product",
            )
        return EOQCPartialCheckResult(
            status=False,
            message=f"FAILED : Variable/Group {var_path} is available in product but has None value",
        )


@dataclass
@EOQCFactory.register_eoqc("var_in_range")
class EOQCVarInRange(EOQCVarBase):
    """
     This EOQC checks that the variables values are within range.

    Parameters
     ----------
     id: str
         The identifier of the quality check.
     version: str
         The version of the quality check in format XX.YY.ZZ .
     thematic: str
         Thematic of the check RADIOMETRIC_QUALITY/GEOMETRIC_QUALITY/GENERAL_QUALITY...
     description: str
         Simple description of the check (less than 100 chars)
     precondition: EOQCFormulaEvaluator
         Precondition evaluator
     variables: list[str]
         List of the variables to check
     min: np.float64
         Min of the range
     max: np.float64
         Maximum of the range
     strict: bool
         Perform a strict compare < of an equal one <=

    """

    min: np.float64
    max: np.float64
    strict: bool = False

    def _var_check(self, var_path: str, value: Any, eoproduct: EOProduct) -> EOQCPartialCheckResult:
        if not isinstance(value, EOVariable):
            raise Exception("Target is not an EOVariable")
        if self.strict:
            if self.min < np.float64(value.data.max().compute().item()) < self.max:
                return EOQCPartialCheckResult(
                    status=True,
                    message=f"PASSED: Variable {var_path} is strictly within {self.min} and {self.max}",
                )
            else:
                return EOQCPartialCheckResult(
                    status=False,
                    message=f"FAILED: Variable {var_path} is not strictly within {self.min} and {self.max}",
                )
        else:
            if self.min <= np.float64(value.data.max().compute().item()) <= self.max:
                return EOQCPartialCheckResult(
                    status=True,
                    message=f"PASSED: Variable {var_path} is within {self.min} and {self.max}",
                )
            else:
                return EOQCPartialCheckResult(
                    status=False,
                    message=f"FAILED: Variable {var_path} is not within {self.min} and {self.max}",
                )


@dataclass
@EOQCFactory.register_eoqc("var_count")
class EOQCCountVar(EOQCVarBase):
    """
    This EOQC check that the variables/groups has the number of expected sub elements.

    Parameters
    ----------
    id: str
        The identifier of the quality check.
    version: str
        The version of the quality check in format XX.YY.ZZ .
    thematic: str
        Thematic of the check RADIOMETRIC_QUALITY/GEOMETRIC_QUALITY/GENERAL_QUALITY...
    description: str
        Simple description of the check (less than 100 chars)
    precondition: EOQCFormulaEvaluator
        Precondition evaluator
    variables: list[str]
        List of the variables to check
    expected: int
        Expected sub groups elements

    """

    expected: int

    def _var_check(self, var_path: str, value: Any, eoproduct: EOProduct) -> EOQCPartialCheckResult:

        if not isinstance(value, EOGroup):
            raise Exception("Target is not a group, can't count element")

        if len(value) != self.expected:
            return EOQCPartialCheckResult(
                status=False,
                message=f"FAILED: Expected {self.expected} elements under {var_path}, found {len(value)}",
            )
        return EOQCPartialCheckResult(
            status=True,
            message=f"PASSED: {self.expected} elements found under {var_path}",
        )


@dataclass
@EOQCFactory.register_eoqc("var_notzerosize")
class EOQCNotZeroSizeVar(EOQCVarBase):
    """
    This EOQC check that the variables has data.

    Parameters
    ----------
    id: str
        The identifier of the quality check.
    version: str
        The version of the quality check in format XX.YY.ZZ .
    thematic: str
        Thematic of the check RADIOMETRIC_QUALITY/GEOMETRIC_QUALITY/GENERAL_QUALITY...
    description: str
        Simple description of the check (less than 100 chars)
    precondition: EOQCFormulaEvaluator
        Precondition evaluator
    variables: list[str]
     List of the variables to check


    """

    def _var_check(self, var_path: str, value: Any, eoproduct: EOProduct) -> EOQCPartialCheckResult:
        if not isinstance(value, EOVariable):
            raise Exception("Target is not an EOVariable")

        if value.data.size == 0:
            return EOQCPartialCheckResult(
                status=False,
                message=f"FAILED: {var_path} variable has no data attached",
            )
        return EOQCPartialCheckResult(
            status=True,
            message=f"PASSED: {var_path} variable has data",
        )
