from pathlib import Path

import numpy as np
from jinja2 import Environment, FileSystemLoader

from eopf import EOProduct
from eopf.accessor.memmap_accessors_impl import FixedMemMap, MemMap
from eopf.common.file_utils import AnyPath
from eopf.exceptions.errors import EOProductError
from eopf.product.eo_variable import EOVariable


class L0Writers:
    @staticmethod
    def S3L0Writer(source: EOProduct, url: str | AnyPath) -> None:
        filename = AnyPath.cast(url)
        filename.mkdir(exist_ok=True)

        ispdat = MemMap(
            filename / "ISPData.dat",
            primary_header_length_bytes=6,
            ancillary_header_length_bytes=0,
            packet_length_start_position_bytes=4,
            packet_length_stop_position_bytes=6,
        )
        # verify mandatory variables
        packet_data_length = source["/conditions/packet_data_length"]
        if not isinstance(packet_data_length, EOVariable):
            raise EOProductError("missing variable for l0 writer : packet_data_length")
        pid = source["/conditions/pid"]
        if not isinstance(pid, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        pcat = source["/conditions/pcat"]
        if not isinstance(pcat, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        packet_sequence_count = source["/conditions/packet_sequence_count"]
        if not isinstance(packet_sequence_count, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        service_packet_type = source["/conditions/service_packet_type"]
        if not isinstance(service_packet_type, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        service_packet_subtype = source["/conditions/service_packet_subtype"]
        if not isinstance(service_packet_subtype, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        destination_id = source["/conditions/destination_id"]
        if not isinstance(destination_id, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        coarse_time = source["/coordinates/coarse_time"]
        if not isinstance(coarse_time, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        fine_time = source["/coordinates/fine_time"]
        if not isinstance(fine_time, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        time_status = source["/conditions/time_status"]
        if not isinstance(time_status, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        user_data = source["/measurements/user_data"]
        if not isinstance(user_data, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        fep = source["/conditions/fep"]
        if not isinstance(fep, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        gps_time_days = source["/coordinates/gps_time_days"]
        if not isinstance(gps_time_days, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        gps_time_seconds = source["/coordinates/gps_time_seconds"]
        if not isinstance(gps_time_seconds, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        gps_time_microseconds = source["/coordinates/gps_time_microseconds"]
        if not isinstance(gps_time_microseconds, EOVariable):
            raise EOProductError("missing variable for l0 writer")

        ispdat.build_buffer(np.asarray(packet_data_length.data))

        ispdat.write_key(0, 3, np.asarray(source.attrs["other_metadata"]["packet_version"]), "s_uint8")
        ispdat.write_key(3, 1, np.asarray(source.attrs["other_metadata"]["packet_type"]), "s_uint8")
        ispdat.write_key(4, 1, np.asarray(source.attrs["other_metadata"]["header_flag"]), "s_bool")
        ispdat.write_key(5, 7, np.asarray(pid.data), "uint8")
        ispdat.write_key(12, 4, np.asarray(pcat.data), "uint8")
        ispdat.write_key(16, 2, np.asarray(source.attrs["other_metadata"]["sequence_flag"]), "s_uint8")
        ispdat.write_key(
            18,
            14,
            np.asarray(packet_sequence_count.data),
            "uint16",
        )
        ispdat.write_key(32, 16, np.asarray(packet_data_length.data), "uint8")
        ispdat.write_key(49, 7, np.asarray(source.attrs["other_metadata"]["packet_category"]), "s_uint8")
        ispdat.write_key(56, 8, np.asarray(service_packet_type.data), "uint8")
        ispdat.write_key(
            64,
            8,
            np.asarray(service_packet_subtype.data),
            "uint8",
        )
        ispdat.write_key(72, 8, np.asarray(destination_id.data), "uint8")
        ispdat.write_key(80, 32, np.asarray(coarse_time.data), "uint32")
        ispdat.write_key(112, 24, np.asarray(fine_time.data), "uint32")
        ispdat.write_key(136, 1, np.asarray(time_status.data), "bool")
        ispdat.write_key(144, -1, np.asarray(user_data.data), "var_bytearray")

        ispdat.save_buffer()
        ispdat.reset()

        ispann = FixedMemMap(url + "/ISPAnnotation.dat", fixed_packet_length_bytes=30)
        ispann.build_buffer(np.asarray(packet_data_length.data).size)

        ispann.write_key(0, 144, np.asarray(fep.data), "bytearray")
        ispann.write_key(144, 32, np.asarray(gps_time_days.data), "uint32")
        ispann.write_key(
            176,
            32,
            np.asarray(gps_time_seconds.data),
            "uint32",
        )
        ispann.write_key(
            208,
            32,
            np.asarray(gps_time_microseconds.data),
            "uint32",
        )

        ispann.save_buffer()
        ispann.reset()

        cur_file_path = Path(__file__).absolute()
        templates_folder_path = cur_file_path.parent.joinpath("templates")
        env = Environment(loader=FileSystemLoader(templates_folder_path), autoescape=True)
        templateS3L0 = env.get_template("S03_L0_manifest.xml")
        attrs = {"PRODUCT": source.attrs}
        output = templateS3L0.render(attrs)

        with (filename / "xfdumanifest.xml").open("w") as f:
            f.write(output)

    @staticmethod
    def S1L0Writer(source: EOProduct, url: str | AnyPath) -> None:
        out_dir = AnyPath.cast(url)
        out_dir.mkdir(exist_ok=True)

        # file name reconstruction based on the name of the output file
        import re

        match = re.search(
            r"(S1[ABCD])_(..)_RAW__0S.._(\d{8}T\d{6})_(\d{8}T\d{6})_(\d{6})_(\w{6})",
            out_dir.path,
        )
        if match:
            file_name_pattern = (
                match[1].lower()
                + "-"  # mission
                + match[2].lower()
                + "-"  # mode
                + "raw"
                + "-"
                + "s-{polarisation}"
                + "-"
                + match[3]
                + "-"  # start time
                + match[4]
                + "-"  # stop time
                + match[5]
                + "-"  # abs orbit number
                + match[6].lower()  # data take
            )
        else:
            file_name_pattern = "ISP-{polarisation}"

        for pola in ["hh", "hv", "vv", "vh"]:
            L0Writers.s1_handle_pola_writings(file_name_pattern, out_dir, pola, source)

        cur_file_path = Path(__file__).absolute()
        templates_folder_path = cur_file_path.parent.parent / "store" / "templates"
        env = Environment(loader=FileSystemLoader(templates_folder_path), autoescape=True)
        attrs = {"PRODUCT": source.attrs, "conditions": source["conditions"].attrs}
        if attrs["PRODUCT"]["stac_discovery"]["properties"]["eopf:instrument_mode"] == "SM":
            templateS1L0 = env.get_template("S01SSMRAW_manifest.xml")
        else:
            templateS1L0 = env.get_template("S01_L0_manifest.xml")
        output = templateS1L0.render(attrs)

        with (out_dir / "manifest.safe").open("w") as f:
            f.write(output)

    @staticmethod
    def s1_handle_pola_writings(file_name_pattern: str, out_dir: AnyPath, polarisation: str, source: EOProduct) -> None:
        if L0Writers.exists(source, f"/conditions/{polarisation}/packet_sequence_count_{polarisation}"):
            file_name = file_name_pattern.replace("{polarisation}", polarisation) + ".dat"

            L0Writers.s1_l0_handle_isp_data(file_name, out_dir, polarisation, source)

            L0Writers.s1_l0_handle_isp_annot(file_name_pattern, out_dir, polarisation, source)

            L0Writers.s1_l0_handle_isp_index(file_name_pattern, out_dir, polarisation, source)

    @staticmethod
    def s1_l0_handle_isp_data(
        file_name: str,
        out_dir: AnyPath,
        polarisation: str,
        source: EOProduct,
    ) -> None:
        packet_data_length = source[f"/conditions/{polarisation}/packet_data_length_{polarisation}"]
        if not isinstance(packet_data_length, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        packet_sequence_count = source[f"/conditions/{polarisation}/packet_sequence_count_{polarisation}"]
        if not isinstance(packet_sequence_count, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        coarse_time = source[f"/coordinates/coarse_time_{polarisation}"]
        if not isinstance(coarse_time, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        fine_time = source[f"/coordinates/fine_time_{polarisation}"]
        if not isinstance(fine_time, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        dataword_index = source[f"/coordinates/dataword_index_{polarisation}"]
        if not isinstance(dataword_index, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        space_packet_count = source[f"/conditions/{polarisation}/space_packet_count_{polarisation}"]
        if not isinstance(space_packet_count, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        pri_count = source[f"/conditions/{polarisation}/pri_count_{polarisation}"]
        if not isinstance(pri_count, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        error_flag = source[f"/quality/error_flag_{polarisation}"]
        if not isinstance(error_flag, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        baq_mode = source[f"/conditions/{polarisation}/baq_mode"]
        if not isinstance(baq_mode, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        baq_length = source[f"/conditions/{polarisation}/baq_length"]
        if not isinstance(baq_length, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        range_decimation = source[f"/conditions/{polarisation}/range_decimation"]
        if not isinstance(range_decimation, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        rx_gain = source[f"/conditions/{polarisation}/rx_gain"]
        if not isinstance(rx_gain, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        tx_ramp_rate = source[f"/conditions/{polarisation}/tx_ramp_rate"]
        if not isinstance(tx_ramp_rate, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        tx_pulse_start_freq = source[f"/conditions/{polarisation}/tx_pulse_start_freq"]
        if not isinstance(tx_pulse_start_freq, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        tx_pulse_length = source[f"/conditions/{polarisation}/tx_pulse_length"]
        if not isinstance(tx_pulse_length, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        rank = source[f"/conditions/{polarisation}/rank"]
        if not isinstance(rank, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        pri = source[f"/conditions/{polarisation}/pri"]
        if not isinstance(pri, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        swst = source[f"/conditions/{polarisation}/swst"]
        if not isinstance(swst, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        swl = source[f"/conditions/{polarisation}/swl"]
        if not isinstance(swl, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        ssb_message = source[f"/conditions/{polarisation}/ssb_message"]
        if not isinstance(ssb_message, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        number_of_quads = source[f"/conditions/{polarisation}/number_of_quads"]
        if not isinstance(number_of_quads, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        user_data = source[f"/measurements/user_data_{polarisation}"]
        if not isinstance(user_data, EOVariable):
            raise EOProductError("missing variable for l0 writer")

        ispdat = MemMap(
            out_dir / file_name,
            primary_header_length_bytes=6,
            ancillary_header_length_bytes=0,
            packet_length_start_position_bytes=4,
            packet_length_stop_position_bytes=6,
        )
        ispdat.build_buffer(np.asarray(packet_data_length.data))
        ispdat.write_key(0, 3, np.asarray(source.attrs["other_metadata"]["packet_version"]), "s_uint8")
        ispdat.write_key(3, 1, np.asarray(source.attrs["other_metadata"]["packet_type"]), "s_bool")
        ispdat.write_key(
            4,
            1,
            np.asarray(source.attrs["other_metadata"]["header_flag"]),
            "s_bool",
        )
        ispdat.write_key(
            5,
            7,
            np.asarray(source.attrs["other_metadata"]["application_process_identifer"]),
            "s_uint8",
        )
        ispdat.write_key(12, 4, np.asarray(source.attrs["other_metadata"]["packet_category"]), "s_uint8")
        ispdat.write_key(16, 2, np.asarray(source.attrs["other_metadata"]["sequence_flag"]), "s_uint8")
        ispdat.write_key(
            18,
            14,
            np.asarray(packet_sequence_count.data),
            "uint16",
        )
        ispdat.write_key(
            32,
            16,
            np.asarray(packet_data_length.data),
            "uint8",
        )
        ispdat.write_key(48, 32, np.asarray(coarse_time.data), "uint32")
        ispdat.write_key(80, 16, np.asarray(fine_time.data), "uint16")
        ispdat.write_key(
            96,
            32,
            np.asarray(source.attrs["other_metadata"]["synchronisation_marker"]),
            "s_uint32",
        )
        ispdat.write_key(
            128,
            32,
            np.asarray(source.attrs["stac_discovery"]["properties"]["eopf:data_take_id"]),
            "s_uint32",
        )
        ispdat.write_key(160, 8, np.asarray(source.attrs["other_metadata"]["event_control_code"]), "s_uint8")
        ispdat.write_key(169, 3, np.asarray(source.attrs["other_metadata"]["test_mode"]), "s_uint8")
        ispdat.write_key(
            172,
            4,
            np.asarray(source.attrs["other_metadata"][f"{polarisation}_receive_channel_id"]),
            "s_uint8",
        )
        ispdat.write_key(
            176,
            32,
            np.asarray(source.attrs["other_metadata"]["instrument_configuration_id"]),
            "s_uint32",
        )
        ispdat.write_key(
            208,
            8,
            np.asarray(dataword_index.data),
            "uint8",
        )
        ispdat.write_key(
            232,
            32,
            np.asarray(space_packet_count.data),
            "uint32",
        )
        ispdat.write_key(264, 32, np.asarray(pri_count.data), "uint32")
        ispdat.write_key(296, 1, np.asarray(error_flag.data), "bool")
        ispdat.write_key(299, 5, np.asarray(baq_mode.data), "uint8")
        ispdat.write_key(304, 8, np.asarray(baq_length.data), "uint8")
        ispdat.write_key(
            320,
            8,
            np.asarray(range_decimation.data),
            "uint8",
        )
        ispdat.write_key(328, 8, np.asarray(rx_gain.data), "uint8")
        ispdat.write_key(
            336,
            16,
            np.asarray(tx_ramp_rate.data),
            "uint16",
        )
        ispdat.write_key(
            352,
            16,
            np.asarray(tx_pulse_start_freq.data),
            "uint16",
        )
        ispdat.write_key(
            368,
            24,
            np.asarray(tx_pulse_length.data),
            "uint32",
        )
        ispdat.write_key(395, 5, np.asarray(rank.data), "uint8")
        ispdat.write_key(400, 24, np.asarray(pri.data), "uint32")
        ispdat.write_key(424, 24, np.asarray(swst.data), "uint32")
        ispdat.write_key(448, 24, np.asarray(swl.data), "uint32")
        ispdat.write_key(
            472,
            48,
            np.asarray(ssb_message.data),
            "bytearray",
        )
        ispdat.write_key(
            520,
            16,
            np.asarray(number_of_quads.data),
            "uint16",
        )
        ispdat.write_key(
            544,
            -1,
            np.asarray(user_data.data),
            "var_bytearray",
        )
        ispdat.save_buffer()
        ispdat.reset()

    @staticmethod
    def s1_l0_handle_isp_annot(
        file_name_pattern: str,
        out_dir: AnyPath,
        polarisation: str,
        source: EOProduct,
    ) -> None:
        packet_data_length = source[f"/conditions/{polarisation}/packet_data_length_{polarisation}"]
        if not isinstance(packet_data_length, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        sensing_time_days = source[f"/coordinates/sensing_time_days_{polarisation}"]
        if not isinstance(sensing_time_days, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        sensing_time_millisec = source[f"/coordinates/sensing_time_millisec_{polarisation}"]
        if not isinstance(sensing_time_millisec, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        sensing_time_microsec = source[f"/coordinates/sensing_time_microsec_{polarisation}"]
        if not isinstance(sensing_time_microsec, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        sensing_time_microsec = source[f"/coordinates/sensing_time_microsec_{polarisation}"]
        if not isinstance(sensing_time_microsec, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        downlink_time_days = source[f"/coordinates/downlink_time_days_{polarisation}"]
        if not isinstance(downlink_time_days, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        downlink_time_millisec = source[f"/coordinates/downlink_time_millisec_{polarisation}"]
        if not isinstance(downlink_time_millisec, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        downlink_time_microsec = source[f"/coordinates/downlink_time_microsec_{polarisation}"]
        if not isinstance(downlink_time_microsec, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        frames = source[f"/conditions/{polarisation}/frames"]
        if not isinstance(frames, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        missing_frames = source[f"/conditions/{polarisation}/missing_frames"]
        if not isinstance(missing_frames, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        crc_flag = source[f"/quality/crc_flag_{polarisation}"]
        if not isinstance(crc_flag, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        vcid_present_flag = source[f"/conditions/{polarisation}/vcid_present_flag"]
        if not isinstance(vcid_present_flag, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        vcid = source[f"/conditions/{polarisation}/vcid"]
        if not isinstance(vcid, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        channel = source[f"/conditions/{polarisation}/channel"]
        if not isinstance(channel, EOVariable):
            raise EOProductError("missing variable for l0 writer")

        file_name = file_name_pattern.replace("{polarisation}", polarisation) + "-annot.dat"
        ispann = FixedMemMap(out_dir / file_name, fixed_packet_length_bytes=26)
        ispann.build_buffer(np.asarray(packet_data_length.data).size)
        ispann.write_key(
            0,
            16,
            np.asarray(sensing_time_days.data),
            "uint16",
        )
        ispann.write_key(
            16,
            32,
            np.asarray(sensing_time_millisec.data),
            "uint32",
        )
        ispann.write_key(
            48,
            16,
            np.asarray(sensing_time_microsec.data),
            "uint16",
        )
        ispann.write_key(
            64,
            16,
            np.asarray(downlink_time_days.data),
            "uint16",
        )
        ispann.write_key(
            80,
            32,
            np.asarray(downlink_time_millisec.data),
            "uint32",
        )
        ispann.write_key(
            112,
            16,
            np.asarray(downlink_time_microsec.data),
            "uint16",
        )
        ispann.write_key(
            128,
            16,
            np.asarray(packet_data_length.data),
            "uint16",
        )
        ispann.write_key(144, 16, np.asarray(frames.data), "uint16")
        ispann.write_key(
            160,
            16,
            np.asarray(missing_frames.data),
            "uint16",
        )
        ispann.write_key(176, 8, np.asarray(crc_flag.data), "bool")
        ispann.write_key(
            184,
            1,
            np.asarray(vcid_present_flag.data),
            "bool",
        )
        ispann.write_key(186, 6, np.asarray(vcid.data), "uint8")
        ispann.write_key(192, 2, np.asarray(channel.data), "s_uint8")
        ispann.save_buffer()
        ispann.reset()

    @staticmethod
    def s1_l0_handle_isp_index(file_name_pattern: str, out_dir: AnyPath, polarisation: str, source: EOProduct) -> None:
        date_time = source[f"/conditions/index/{polarisation}/date_time"]
        if not isinstance(date_time, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        date_time = source[f"/conditions/index/{polarisation}/date_time"]
        if not isinstance(date_time, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        delta_time = source[f"/conditions/index/{polarisation}/delta_time"]
        if not isinstance(delta_time, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        delta_size = source[f"/conditions/index/{polarisation}/delta_size"]
        if not isinstance(delta_size, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        data_offset = source[f"/conditions/index/{polarisation}/data_offset"]
        if not isinstance(data_offset, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        byte_offset = source[f"/conditions/index/{polarisation}/byte_offset"]
        if not isinstance(byte_offset, EOVariable):
            raise EOProductError("missing variable for l0 writer")
        var_size_flag = source[f"/conditions/index/{polarisation}/var_size_flag"]
        if not isinstance(var_size_flag, EOVariable):
            raise EOProductError("missing variable for l0 writer")

        file_name = file_name_pattern.replace("{polarisation}", polarisation) + "-index.dat"
        ispidx = FixedMemMap(out_dir / file_name, fixed_packet_length_bytes=36)
        ispidx.build_buffer(np.asarray(date_time.data).size)
        ispidx.write_key(
            0,
            64,
            np.asarray(date_time.data),
            "uint64",
        )
        ispidx.write_key(
            64,
            64,
            np.asarray(delta_time.data),
            "uint64",
        )
        ispidx.write_key(
            128,
            32,
            np.asarray(delta_size.data),
            "uint32",
        )
        ispidx.write_key(
            160,
            32,
            np.asarray(data_offset.data),
            "uint32",
        )
        ispidx.write_key(
            192,
            64,
            np.asarray(byte_offset.data),
            "uint64",
        )
        ispidx.write_key(
            256,
            8,
            np.asarray(var_size_flag.data),
            "bool",
        )
        ispidx.save_buffer()
        ispidx.reset()

    # TODO move to EOProduct directly !!!
    @staticmethod
    def exists(product: EOProduct, the_path: str) -> bool:
        try:
            product[the_path]
            return True
        except KeyError:
            return False
